***********************************************************************
* Signal Extension
*  Richard Zidlicky
***********************************************************************
*          Changes
*          ^^^^^^^
* v 0.12 :
*          - first testing done by KP Greiner
*          - included  with APL11 interpreter
* v 0.24 :
*          - first real release
*          - added timers
*          - added %CSG handler to get Unix like behaviour easier 
*          - added sigvec interface for maintaining %CSG handler,
*            sigcleanup(),sigraise(),sigaction(),sigprocmask(),sigpending()
*          - send/receive message now basic mechanism for communication
*            with signal device (io_sstrg and io_edlin traps)
*          - changed signal numbering in signal.h to match at least
*            some of the signals defined in X/OPEN 87
* v 0.25 :
*          - fixed bug in sub_penq
* v 0.26 :
*          - fixed bug in insert_event, reported by Phil Borman
* v 0.27 :
*          - rewrite of checksig()/raise()
*          - bug fixes for defering
*          - added system default handlers/table
*          - preparing new features, more error checking
* v 0.30 :
*          - removed some experimental features, mainly ring protection
*            scheme
*          - added emulated unix system call handling and coresponding
*            sigvec calls
*          - changed sigvec calls to always apply to job's own sighandler,
*            usefull for threads and shared libraries
*          - added message to allocate and init handler '%CSG'
*
***********************************************************************
* structure of signal handler : simple case
*  $00     word        $4afc
*  $02     long        '%SIG'      magic
*  $06     long        sighnd      pointer to handler routine or NULL

*  $0a     long        stackmin    no interrupt if sp < stackmin                   
*  $0e     long        priority    
***

* and the coresponding defines

h_magic    equ         0
h_type     equ         2
*

hs_handler equ         6
hs_stackmin equ        $a
hs_prio    equ         $e
hs_len     equ         18

*********************
* structure of signal handler : extended handler
*  $00     word        $4afc                   
*  $02     long        '%CSG'          magic
*  $06     long        MAXSIG          number of signals handled
*          long        *sigcurrent[MAXSIG+1]  ptr to array of handlers
*          long        *sigdefault[MAXSIG+1]  NULL or .. 
*          long        *sa_mask[MAXSIG+1]       sigs 2be blckd while handling
*          word        void                    * was hprot_id
*          long        void                    * was hprotlevel([])
*          word        stack_id
**                                 0 : lowstacklimit
**                                 ni 1 : *{sttop;stbottom} ; 1 xtra stack
**                                 ni 2 : *{sttop;stbottom}[MAXSIZ+1]
*          long        (*)stackstruct([])       lowstacklimit or extra stacks
*          word        pri_id
**                                 0 : priority
**                                 1 : *priority[MAXSIZ+1]
*          long        (*)pristruct([])         priority(ies)
*          long        *scinf           system call emulation stuff
*          long        (*sigvec)()     pointer to interface routine
*
*** some status info 
*          word        void        * was cprotlevel
*
*          byte        deffer      intern flag, defered signal
*          byte        ixx         intern flag, checksig in progress
*
*          word        activ       set when signalhandling, cleared by signal(),
*                                  sigaction(), sigreturn()
*                                  to avoid handler beeing reset to old
*          long        uval        NULL or *uval[nsigs+1], stored# 
*          long        block                    block this signals
*          long        pending                  mask of # signals
*          long        resethandler             # if handling 1of these ..
*** 
* and its defines

hc_nsigs   equ         6           .l
hc_curr    equ         10          .l
hc_def     equ         14          .l
hc_samask  equ         18          .l
*hc_protid equ         22          .w
*hc_protlev equ        24          .l
hc_stackid equ         28          .w
hc_stack   equ         30          .l
hc_priid   equ         34          .w
hc_prio    equ         36          .l
hc_scinf   equ         40          .l
hc_sigvec  equ         44          .l
*hc_chlev  equ         48          .w          * unused
hc_defer   equ         50          .b
hc_ixx     equ         51          .b
hc_activ   equ         52          .w
hc_uval    equ         54          .l
hc_block   equ         58          .l
hc_pend    equ         62          .l
hc_reset   equ         66          .l
hc_len     equ         70
** total 70 **

***********************************************************************

* messages for communication with signal device
*
* internal link block
ml.nxt     equ         4           ; 
ml         equ         8          ; link block size 

* message header definition
msg.magic  equ         0           ; '%MSG'
msg.len    equ         4
msg.type   equ         8           ; message type
*
MT_SIGNAL  equ         0
MT_TIMER   equ         1           
MT_HINIT   equ         2           ; alloc and init handler

msg.txi    equ         10          ; .w  timer info (unit),0=ticks,1=secs
*                                  ;  hinit: HI_PR | HI_UV | HI_SYSC
msg.jobid  equ         12          ; .l

* for signal and timer types
msg.signr  equ         16          ; .l
msg.prio   equ         20          ; .l
msg.uval   equ         24          ; .l
*
sigmsg_len equ         28

msg.t_evid equ         28          ; .l  timer event id, 0 reserved, 1=alarm, 
msg.t_ticks equ        32          ; .l  send signal after ticks units
msg.t_int  equ         36          ; .l  interval timer if <>0
tmsg_len   equ         40

* for hinit command
msg.hi_nsigs equ       16          ; nr of signals to handle - 0== simple handler
msg.hi_stack equ       20          ; reserve some place
himsg_len  equ         24

msg_maxlen equ         tmsg_len
msg_minlen equ         himsg_len

*

SIG_IGN    equ         0
SIG_DFL    equ         1

* priority:
WFJOB      equ         %111000000000  ; mask interrupt if waiting for job
SUSP       equ         %000000111000  ;                              trap#1
WFIO       equ         %000111000000  ;                              trap#3
NORM       equ         %000000000111  ;        normal execution

* optional flags for:
*   priority_s:
DF_WIO     equ         %10000000000000  ; defer if waitiong for io
DF_SUSP    equ         %1000000000000   ;          suspended
DF_WJOB    equ         %100000000000000 ;          waiting for job
*SIGPG     equ         %1000000000000000 ; send to process group

*   priority_r:
*PGROOT    equ        %1000000000000   ; root of process group

* priority: mask of reserved bits
P_RES_R    equ    ^(NORM+SUSP+WFIO+WFJOB)      
P_RES_S    equ    ^(DF_WIO+DF_SUSP+DF_WJOB+NORM+SUSP+WFIO+WFJOB) 

***********
* stackframe of saved context (struct SIGSVB )
* used for %SIG handler
svb.finish equ         $0          ; pointer to interrupt cleanup routine,
*                                    usualy called by rts
svb.sgn    equ         $4          ; .W : 0 normal, 1 io-susp, 2 waitforjob
*                                  *      3 susp
sgn_nrm    equ         0           ; normal
sgn_ios    equ         1           ; io suspension
sgn_wfj    equ         2           ; wait for job
sgn_sus    equ         3           ; suspension
sgn_raise  equ         4           ; raise() ed signal
* rest of structure is undefined for raised signals


svb.fs1    equ         $6          ; 2 svb.sgn specific longwords
svb.fs2    equ         svb.fs1+4

svb.d4     equ         svb.fs2+4   ; saved registers
svb.d5     equ         svb.d4+4
svb.d6     equ         svb.d5+4
svb.a4     equ         svb.d6+4
svb.a5     equ         svb.a4+4
svb.a7     equ         svb.a5+4

svb.sr     equ         svb.a7+4
svb.ret    equ         svb.sr+2

svb_len    equ         svb.ret+4


*** extended handler sigcontext ***
* struct sigcontext / esvb used for %CSG handlers
* 
* svb.*                                        ; same as above
esvb.oblock equ        svb_len                 ; block mask to restore
esvb.ohand equ         esvb.oblock+4           ; handler to restore or 0
esvb.signr equ         esvb.ohand+4
esvb.hstruct equ       esvb.signr+4            ; pointer to handler struct
esvb.uval  equ         esvb.hstruct+4          
esvb.d0    equ         esvb.uval+4 ; saved compiler scratch registers
esvb.d1    equ         esvb.d0+4
esvb.d2    equ         esvb.d1+4
esvb.a0    equ         esvb.d2+4
esvb.a1    equ         esvb.a0+4
esvb_len   equ         esvb.a1+4

* shadow definitions for cint_ret hack
esvb.ret   equ         esvb_len-4
esvb.sr    equ         esvb.ret-2

***********************************

* temporary storage block for cseval and do_int
tmp.htype  equ         0                       ; type of handler
tmp.haddr  equ         tmp.htype+4 ; ptr handler routine or cntxt
tmp.stack  equ         tmp.haddr+4 ; stackbottom
tmp.prio   equ         tmp.stack+4
tmp.imode  equ         tmp.prio+4              ; interrupt mode (syscall)
tmp.force  equ         tmp.imode+2 ; syscall mode
tmp.sregs  equ         tmp.force+2 ; flag, ignore blocking

tmp_size   equ         tmp.sregs+3*4           ; space for 3 saved regs

* tmp.imode :
IM_NORM    equ         0           ; normal interrupt
IM_EXP     equ         1           ; make QDOS calls expire
IM_BLOCK   equ         2           ; simply block it
IM_LJMP    equ         3           ; longjmp to tmp.haddr=si_retcntx


**
* emulated system call stuff
*
* struct SYSCE, 
*
*  word    stat        SCS_NOSYS,SCS_ATOM,SCS_EXP
*  long    *link                   link to higher level SYSCE
*  long    *pend                   signals received here
*
*  long    a3-a7,PC                longjmp context

*sce.id    equ         0
sce.stat   equ         2-2
sce.nx     equ         4-2
sce.pend   equ         8-2

sce.cntxt  equ         12-2
sce.a3     equ         sce.cntxt
sce.a4     equ         sce.a3+4
sce.a5     equ         sce.a3+8
sce.a6     equ         sce.a3+12
sce.a7     equ         sce.a6+4
sce.pc     equ         sce.a7+4
sce_size   equ         sce.pc+4
*
SCS_NOSYS  equ         0
SCS_EXP    equ         1
SCS_ATOM   equ         2
*


******************************************************************
*
* QDOS system definitions
*

TRAP3      equ         $4e43       ; opcode
SUPER      equ         $2000       ; s-bit in sr

** signals that need special attention

SIGILL     equ         4
SIGSEGV    equ         6
SIGBUS     equ         7
_SIGFTX    equ         (1<<(SIGILL-1))+(1<<(SIGSEGV-1))+(1<<(SIGBUS-1))

SIGKILL    equ         9
*SIGTERM   equ         15

NOBLOCKMASK equ        ^(1<<(SIGKILL-1))       ; can't block this
NOCATCH    equ         SIGKILL


EINTR      equ         -1          ; frjob return code

sgn_def    equ         100         ; intern define

*   Trap #1 keys

sms.info   equ         $00         ; get system INFOrmation
sms.achp   equ         $18         ; Allocate Commaon HeaP
mt.alchp   equ         sms.achp
sms.lpol   equ         $1c         ; Link in POLled action
sms.lshd   equ         $1e         ; Link in SCheDuler action
sms.liod   equ         $20         ; Link in IO Device driver

*   Vectors

mm.alchp   equ         $c0
mm.rechp   equ         $c2
ut.wtext   equ         $d0         ; Write TEXT vector


*   System Variables

sv_jbpnt   equ         $64         ; current JoB PoinTer
sv_jbbas   equ         $68         ; JoB Table Base
sv_jbtop   equ         $6c         ; JoB Table Top
sv_chmax   equ         $72
sv_chbas   equ         $78         ; CHannel Table Base
sv_chtop   equ         $7c         ; CHannel Table Top


*   Jobheader Offsets

job.stat   equ         $14
job.tag    equ         $10
job.wflag  equ         $17
job.wjob   equ         $18
job.d0     equ         $20
job.d1     equ         job.d0+4
job.d2     equ         job.d1+4
job.d3     equ         job.d2+4
job.d4     equ         job.d3+4
job.d5     equ         job.d4+4
job.d6     equ         job.d5+4

job.a0     equ         $40
job.a1     equ         job.a0+4
job.a2     equ         job.a1+4
job.a3     equ         job.a2+4
job.a4     equ         job.a3+4
job.a5     equ         job.a4+4
job.a6     equ         job.a5+4

job.pc     equ         $62
job.sr     equ         $60
job.sp     equ         $40+7*4
job.a7     equ         job.sp


*   Signal device offsets

dev_plblk  equ         $8                      ; poll link block
dev_pladr  equ         $c
dev_scblk  equ         $10                     ; sched link block
dev_scadr  equ         $14

dev_link   equ         $18
dev_io     equ         $1c         
dev_open   equ         $20         
dev_close  equ         $24         
dev_joblist equ        dev_close+4           ; ll of signal-recving channels
dev_nxevent equ        dev_joblist+4           ; ptr to next timer event
dev_nticks equ         dev_nxevent+4           ; #ticks till next event
dev_gticks equ         dev_nticks+4            ; global tick count
dev_scratch equ        dev_gticks+4

dev_size   equ         dev_scratch+48

* trap #3 interface
SEND.SIG   equ         $2f                     ; old call
SEND.MSG   equ         $7
SR.MSG     equ         $4
INFO       equ         $3

UCMASK0    equ         $ffdfdfdf
UCMASK1    equ         $dfdfdfff


*  Channel block definition

ch_driver  equ         $4
ch_jobwt   equ         $14
ch_actn    equ         $13
ch_tag     equ         $10
ch_stat    equ         $12
ch_ojob    equ         $08

* private variables in channel definition

ch_nxtch   equ         $14+4                   ; next receiving in ll 
ch_sigh    equ         ch_nxtch+4              ; pointer to handler struct
ch_events  equ         ch_sigh+4               ; ll of timer events
ch_ispf    equ         ch_events+4 ; scheduler loop flag
ch_alloc   equ         ch_ispf+2

ch_len     equ         ch_alloc+2

ERR.NC     equ         -1
ERR.NJ     equ         -2
ERR.OM     equ         -3
ERR.OR     equ         -4
ERR.NO     equ         -6
ERR.NF     equ         -7
ERR.IU     equ         -9
ERR.BN     equ         -12
ERR.BP     equ         -15
ERR.XP     equ         -17
ERR.NI     equ         -19
ERR.BL     equ         -21
  


***********************************************************
* dev init code

         SECTION     TEXT

install:
           moveq       #dev_size,d1
           moveq       #sms.achp,d0            
           clr.l       d2
           trap        #1
           tst.l       d0
           bne.s       exit_ini
           move.l      a0,a4

           lea         sched,a1
           move.l      a1,dev_scadr(a0)
           lea         poll,a1
           move.l      a1,dev_pladr(a0)
           lea         open,a1
           move.l      a1,dev_open(a0)
           lea         io,a1
           move.l      a1,dev_io(a0)
           lea         close,a1
           move.l      a1,dev_close(a0)
           clr.l       dev_joblist(a0)         ; (-$18) init joblist   
           clr.l       dev_nxevent(a0)
           clr.l       dev_nticks(a0)

           lea         dev_link(a0),a0
           moveq       #sms.liod,d0
           trap        #1
           tst.l       d0
           bne.s       exit_ini
           lea         dev_scblk(a4),a0
           moveq       #sms.lshd,d0
           trap        #1
           lea         dev_plblk(a4),a0
           moveq       #sms.lpol,d0
           trap        #1

           suba.l      a0,a0
           lea         init_text,a1
           move.w      ut.wtext,a2
           jsr         (a2)
exit_ini:
           rts

version    equ         '0.30'

init_text: dc.w        ite-itsst
itsst      dc.b        'Signal extension v0.30, '
           dc.b        ' Richard Zidlicky',10
ite:
           ds.w      0
           

***********************************************************
* devdriver code

******************
* poll routine : dev_gticks++, if dev_nticks!=0 then  dev_nticks--
poll:
           addq.l      #1,dev_gticks(a3)
           tst.l       dev_nticks(a3)
           beq.s       rtpoll
           subq.l      #1,dev_nticks(a3)
rtpoll:    rts

******************
* sched routine : send signal if dev_nticks==0 && dev_nxevent!=0
sched:
****       movem.l     d3-d7/a3-a5,-(a7)
rsched:    tst.l       dev_nticks(a3)
           bne.s       rsc
           tst.l       dev_nxevent(a3)
           beq.s       rsc
           move.l      dev_nxevent(a3),a0
           addq.l      #ml,a0

           move.l      #'%MSG',msg.magic(a0)
           clr.w       msg.type(a0)            ; just a signal
           move.l      a0,a4                   ; arg for _send_msg
           suba.l      a5,a5
           moveq       #1,d3

           bsr         _send_msg               ; ... send the signal

           cmpa.l      #0,a5
           beq         scerr_spur              ; pretty bad ...

           addq.w      #MT_TIMER,msg.type(a0)
           subq.l      #ml,a0

           move.l      ml.nxt(a0),ch_events(a5)            ; unlink
           move.l      msg.t_int+ml(a0),d0                 ; interval timer ?
           beq.s       scnoint

           move.l      d0,msg.t_ticks+ml(a0)   ; new tick count
           move.l      a0,a4
           clr.w       d7                      ; no response !
           bsr         insert_event            ; insert time sorted
scfe:      bsr         find_next_event         ; reschedule
           bra.s       rsched                  ; more pending events?


scnoint:   move.l      a3,a4
           move.w      mm.rechp,a2
           jsr         (a2)
           move.l      a4,a3
           bra.s       scfe

* now check for channels that require expiring QDOS calls
rsc:       move.l      dev_joblist(a3),d0
           beq.s       rtss
rscl:      move.l      d0,a5
           tst.w       ch_ispf(a5)
           beq.s       rsclnx
           bsr.s       qexpire
rsclnx     move.l      ch_nxtch(a5),d0
           bne.s       rscl
rtss:
*****      movem.l     (a7)+,d3-d7/a3-a5
           rts

qexpire:   subq.w      #1,ch_ispf(a5)
           bne.s       rtsc
           addq.w      #1,ch_ispf(a5)

           lea         0-tmp_size(a7),a7
           move.l      a7,a2                   ; alloc tmp struct

           move.l      ch_sigh(a5),a4
           cmpi.l      #'%CSG',h_type(a4)
           bne.s       uqx
           move.l      hc_scinf(a4),d0
           beq.s       uqx
           move.l      d0,a1
           cmp.w       #SCS_EXP,sce.stat(a1)
           bne.s       uqx
           move.l      sce.pend(a1),a1
           tst.l       (a1)
           beq.s       uqx

           move.l      ch_ojob(a5),d1
           move.l      d1,d4

           move.l      #%01001001,tmp.prio(a2) ; fake, priority
           move.l      #%11011011,d6             ;    already tested

           bsr         getjbasea0
           tst.w       d0                      ; debug only, shouldn't happen
           bne.s       uqbj

           move.l      hc_scinf(a4),a1
           move.l      job.a7(a0),d0           ; check stack, debug only
           cmp.l       sce.a7(a1),d0
           bcc.s       uqbs

           bsr         sub_penq
           tst.w       d0
           ble.s       urqx                    ; error or nothing to do
           bsr         sub_sysci
           addq.w      #4,ch_ispf(a5)          ; avoid infinite timeout loops
           bra.s       urqx

uqx:       clr.w       ch_ispf(a5) ; condition gone
urqx:      lea         tmp_size(a7),a7
rtsc:      rts

uqbj:
           suba.l      a0,a0
           clr.l       dev_nxevent(a3)
           lea         sbj_txt,a1
           move.w      ut.wtext,a2
           jsr         (a2)
           bra.s       urqx

sbj_txt:   dc.w        sbje-sbjs
sbjs:      dc.b        'signal extension: bad job in scheduler loop'
sbje       ds.w      0

uqbs:
           suba.l      a0,a0
           clr.l       dev_nxevent(a3)
           lea         sbs_txt,a1
           move.w      ut.wtext,a2
           jsr         (a2)
           bra.s       urqx

sbs_txt:   dc.w        sbse-sbss
sbss:      dc.b        'signal extension: bad stack in scheduler loop'
sbse       ds.w      0

* should not really happen, uncomment for debuging:
scerr_spur:
           clr.l       dev_nxevent(a3)         
**         bra         rsc
           suba.l      a0,a0
           clr.l       dev_nxevent(a3)
           lea         scerr_txt,a1
           move.w      ut.wtext,a2
           jsr         (a2)
           rts
scerr_txt: dc.w        scete-scets
scets:     dc.b        'signal extension:  timer ll corruption',10
scete      ds.w      0

getjbasea0:
           moveq       #0,d0
           move.w      d1,d0
           asl.l       #2,d0
           move.l      sv_jbbas(a6),a0
           adda.l      d0,a0
           moveq       #ERR.NJ,d0
           cmpa.l      sv_jbtop(a6),a0
           bcc.s       gjb_bj
           tst.b       (a0)                    ; job removed ?
           bmi.s       gjb_bj
           move.l      (a0),a0                 ; get jobheader
           swap        d1
           cmp.w       job.tag(a0),d1
           bne.s       gjb_bj
           swap        d1
           clr.l       d0
gjb_bj     rts


******************************************
open:
           move.l      a0,a2
           move.l      d1,d7                   ; save jobid
           move.w      (a2)+,d2                ; check name
           move.l      (a2)+,d1
           andi.l      #UCMASK0,d1 
           cmp.l       #'*SIG',d1
           bne.s       err_nf
           move.l      (a2)+,d1
           andi.l      #UCMASK1,d1 
           cmp.l       #'NAL_',d1
           bne.s       err_nf
           bra.s       nm_found
           dc.w        6
           dc.b        'SIGNAL'

nm_found:  subq.w      #8,d2
           beq.s       err_bp
           blt.s       err_nf
           subq.w      #1,d2
           bne.s       err_bn

           move.b      (a2)+,d1
           andi.b      #$df,d1

           cmp.b       #'S',d1                 ; send only
           beq.s       o_send_sig
           cmp.b       #'R',d1                 ; send & receive
           beq.s       o_sr_sig                ; init signal handler

err_bn:    moveq       #-12,d0
           rts
err_nf:    moveq       #-7,d0
           rts
err_bp:    moveq       #-15,d0
           rts
err_iu:    move.w      mm.rechp,a2
           jsr         (a2)                    ; release memory
           moveq       #-9,D0
           rts

* open send/receive
o_sr_sig:
           move.l      d3,d5                   ; save ihptr
           beq.s       o_send_sig              ; no address, maybe later
           andi.b      #1,d3
           bne.s       err_bp                  ; uneven address
           move.l      a3,a5                   ; save
           move.w      mm.alchp,a2
           move.l      #ch_len,d1
           jsr         (a2)
           tst.l       d0
           bne.s       rtfo

           lea         dev_joblist-ch_nxtch(a5),a1         ; adjust to "previous"
loop:      move.l      ch_nxtch(a1),d0
           beq.s       eloop
           move.l      d0,a1
           cmp.l       ch_ojob(a1),d7          ; max one handler per job!
           beq.s       err_iu
           bra.s       loop
eloop:     move.l      a0,ch_nxtch(a1)

init_ch    move.l      d5,ch_sigh(a0)
*          clr.l       ch_nxtch(a0)
*          clr.l       ch_events(a0)
*          clr.w       ch_ispf(a0) ; no pending signals
           moveq       #0,d0

** now check whether handler is '%csg', if so init some values
           move.l      d5,a1
           cmp.w       #$4afc,(a1)
           bne.s       rtfo                    ; should really give some warning
           cmp.l       #'%CSG',h_type(a1)
           bne.s       rtfo
           lea         c_sigvec,a2 ; vector for clib calls
           move.l      a2,hc_sigvec(a1)
*          clr.l       hc_block(a1)            ; mm.alchp clears it
*          clr.l       hc_pend(a1)
*          clr.w       hc_defer(a1)

rtfo:      rts

* open send only
o_send_sig:
           move.l      #ch_len,d1
           move.w      mm.alchp,a2
           jsr         (a2)
*          clr.l       ch_sigh(a0) ; mm.alchp clears it
*          clr.l       ch_nxtch(a0)
*          clr.l       ch_events(a0)
           bra.s       rtfo


******
close:
           movem.l     d4-d7/a4-a5,-(a7)
           move.l      a0,a5

           tst.l       ch_sigh(a0)
           beq.s       ecls                    ; send only, no unlink needed

           move.l      ch_sigh(a0),a1
           cmp.l       #'%CSG',h_type(a1)
           bne.s       cevnt
           clr.w       hc_activ(a1)            ; still could be inside handler
           clr.l       h_type(a1)

* cancel all ch_events, free their memory
cevnt:
           move.l      ch_events(a0),a4
           cmpa.l      #0,a4
           beq.s       unlink

           clr.l       ch_events(a0)
           move.l      a3,d4

l_rm:      move.l      a4,a0
           move.l      ml.nxt(a0),a4           ; next
           move.w      mm.rechp,a2
           jsr         (a2)                    ; release mem
           cmpa.l      #0,a4
           bne.s       l_rm

cl_erm:    move.l      d4,a3
           move.l      a5,a0

unlink:    move.l      dev_joblist(a3),d1
           beq.s       cret
           lea         dev_joblist-ch_nxtch(a3),a1         ; adjust to "last" ptr
cll:       cmp.l       d1,a0
           beq.s       eclsf
           move.l      d1,a1
           move.l      ch_nxtch(a1),d1
           bne.s       cll

cret       movem.l     (a7)+,d4-d7/a4-a5       
           rts                                 ; not in list ??

eclsf:
           move.l      ch_nxtch(a0),ch_nxtch(a1)           ; unlink channel

           bsr         find_next_event                     ; reschedule events
ecls:
*          tst.w       ch_alloc(a5)
*          beq.s       clna
*          move.l      ch_sigh(a5),a0          ; return handler if allocated
*          suba.l      #12,a0
*          move.w      mm.rechp,a2
*          jsr         (a2)
clna       move.l      a5,a0                   ; return chdef block
           movem.l     (a7)+,d4-d7/a4-a5
           move.w      mm.rechp,a2
           jmp         (a2)                    ; return memory and rts

           

**********
io:        
           cmp.b       #SEND.SIG,d0            ; old entry point?
           beq         send_sig
           moveq.l     #0,d7
           moveq.l     #0,d6
           and.l       #$ffff,d1
           and.l       #$ffff,d2
           cmp.b       #SR.MSG,d0
           beq.s       sr_msg
           move.l      d2,d5                   ; length
           cmp.b       #SEND.MSG,d0
           beq.s       send_msg

           cmp.b       #INFO,d0
           beq.s       info_msg
           moveq       #-15,d0
           rts
err_m_or:  moveq       #-4,d0
           rts

* io_fstrg : write out struct SIG_INFO 
info_msg:
           moveq.l     #12,d0
           move.l      d0,d1
           cmp.w       d0,d2
           bcc.s       iic
           move.l      d2,d1
iic:
           move.l      d1,d0
imc:       lea         dev_scratch(a3),a4
           move.l      #version,(a4)+
           move.l      dev_gticks(a3),(a4)+
           lea         c_sigvec,a5
           move.l      a5,(a4)

           subq.l      #8,a4
           subq.w      #1,d0
imcc:      move.b      (a4)+,(a1)+
           dbra        d0,imcc
           clr.w       d0
           rts

_send_msg:
           movem.l     d1-d2/a0/a1/a3,-(a7)
           bra.s       _sxmg

* io_edlin parameters
sr_msg:    suba.l      d1,a1                   ; start of message
           st          d7                      ; flag: return value wanted
           move.l      d1,d5                   ; actual length

* io_sstrg parameters
send_msg:
           lea         dev_scratch(a3),a4
           moveq       #msg_maxlen-1,d4
           move.l      a1,a2
smcpm:     move.b      (a2)+,(a4)+ ; copy to scratch to insure
           dbra        d4,smcpm                ; word alignment
           
           lea         dev_scratch(a3),a4
           tst.b       d7
           bne.s       _smsg
           move.l      msg.len(a4),d6          ; len for io.sstrg return
_smsg:
           movem.l     d1-d2/a0/a1/a3,-(a7)
           cmp.w       #msg_minlen,d5          ; too short
           bcs         err_nomsg
           move.l      msg.len(a4),d4
           cmp.w       d4,d5                   ; complete msg must be in buffer
           bcs         err_nomsg
           cmp.w       d4,d2
           bcs         err_nomsg

_sxmg:
           cmp.l       #'%MSG',msg.magic(a4)
           bne         err_blmsg
           move.l      msg.jobid(a4),d1
           bpl.s       snn
           bsr         getjobid                ; -1 ->current
           move.l      d1,msg.jobid(a4)

snn:       tst.w       msg.type(a4)            ; signal ?
           bne.s       set_alarm

           movem.l     d6/d7/a1,-(a7)
           move.l      msg.signr(a4),d2
           move.l      msg.jobid(a4),d1
           move.l      msg.uval(a4),a2
           move.l      msg.prio(a4),a1

           bsr         send_sig
           movem.l     (a7)+,d6/d7/a1
           bra         sm_ret

set_alarm:
           cmp.w       #MT_TIMER,msg.type(a4)
           bne         hinit
           cmp.l       #tmsg_len,msg.len(a4)
           bcs         err_nomsg
           moveq       #0,d2                   ; signr := test
           move.l      msg.jobid(a4),d1

           movem.l     d6/d7/a4,-(a7)
           move.l      msg.prio(a4),a1
           bsr         send_sig                ; test job and handler
           movem.l     (a7)+,d6/d7/a4
           move.l      d1,msg.jobid(a4)        ; real jobid if jobid=-1
           movem.l     (a7),d1-d2/a0/a1/a3

* a5= chdefbl for jobid
           cmp.w       #-20,d0
           beq.s       csam                    ; ignore low priority
           tst.w       d0
           bne.s       sm_ret

csam:      subq.l      #ml,a4                  ; point to linkage block
           tst.l       msg.signr+ml(a4)
           beq.s       mscnt                   ; cancel request; dont alloc
           move.l      msg.len+ml(a4),d1
           addq.l      #ml,d1                  ; extra space for link pointer
           move.w      mm.alchp,a2
           jsr         (a2)
           tst.w       d0
           bne.s       sm_ret

           addq.l      #ml,a0
           addq.l      #ml,a4
           move.l      msg.len(a4),d2
           move.l      a4,a1
           move.l      a0,a4                   ; save addr of block
           subq.l      #1,d2
cpymsg:    move.b      (a1)+,(a0)+
           dbra        d2,cpymsg
           subq.l      #ml,a4                  ; point to msg-4

* a4=lmsg, a3=dev, a5=chdef
mscnt:     movem.l     (a7),d1-d2/a0/a1/a3
           bsr         insert_event
           movem.l     (a7),d1-d2/a0/a1/a3
           bsr         find_next_event         ; reschedule events
           clr.l       d0                      ; no error

sm_ret     movem.l     (a7)+,d1-d2/a0/a1/a3
           tst.w       d0
           bne.s       srs
           move.l      d6,d1                   ; msg.len or old msg.len
sr1:       adda.l      d1,a1                   ; behave like a real io.sstrg
sr2:       rts                                 ;   or io.edlin

srs:       tst.w       d7
           bne.s       sr1
           rts

msgerr.ni:
err_msg_ni:
           moveq       #ERR.BN,d0
           bra.s       sm_ret
msgerr.bp:
err_nomsg:
           moveq       #ERR.BP,d0
           bra.s       sm_ret
msgerr.bl:
err_blmsg:
           moveq       #ERR.BL,d0
           bra.s       sm_ret
msgerr.bj:
           moveq       #ERR.NJ,d0
           bra.s       sm_ret
msgerr.iu:
           moveq       #ERR.IU,d0
           bra.s       sm_ret
msgerr.or:
           moveq       #ERR.OR,d0
           bra.s       sm_ret
msgerr.om:
           moveq       #ERR.OM,d0
           bra.s       sm_ret

* alloc & init handler, add to list
* chan a0 must be owned by msg.jobid & must be only one used as handler
*
hinit:     cmp.w       #MT_HINIT,msg.type(a4)
           bne.s       err_msg_ni
           cmp.l       #himsg_len,msg.len(a4)
           bcs.s       err_nomsg
* now check that owner of a0 is msg.jobid
           cmp.l       ch_ojob(a0),d1
           bne.s       msgerr.bj
* check that no handler for this job exists
           move.l      a0,a1
           bsr         fjhand
           tst.l       d0
           bne.s       msgerr.iu               ; job already has an handler
           move.l      a1,a0
* compute space required
           addq.l      #1,msg.hi_nsigs(a4)     ; all arrays dim(NSIG+1) !!!!!
           move.l      msg.hi_nsigs(a4),d0
           beq.s       msgerr.or               ; hook in '%SIG handler'
           bmi.s       msgerr.bp
           cmp.l       #32,d0
           bhi.s       msgerr.or

           moveq.l     #8,d1                   ; x size (current+samask)
           move.w      msg.txi(a4),d2
           beq.s       hnsys                   ; no extras
           btst        #0,d2                   ; HI_UVAL ?
           beq.s       hiuc
           addq.l      #4,d1
hiuc       btst        #1,d2                   ; HI_PRIO ?
           beq.s       hiuc1
           addq.l      #4,d1
hiuc1
           mulu        d0,d1
           add.w       #hc_len,d1              ; add basic size
*          btst        #2,d2                   ; HI_SYSC ?
*          beq.s       hnsys
*          addi.w      #sce_size,d1

hnsys:
           move.l      a0,a5                   ; save channel addr
           movem.l     d2/a1-a3,-(a7)          ; check mask
           move.l      msg.jobid(a4),d2
           moveq.l     #mt.alchp,d0
           trap        #1
           movem.l     (a7)+,d2/a1-a3
           tst.l       d0
           bne.s       msgerr.om
* setup handler
           move.l      a0,a1
           move.w      #$4afc,(a1)+
           move.l      #'%CSG',(a1)+
           move.l      msg.hi_nsigs(a4),(a1)
           subq.l      #1,(a1)     

           move.l      msg.hi_stack(a4),hc_stack(a0)

* now set all arrays and extras
           move.l      a0,a1
           move.l      msg.hi_nsigs(a4),d1
           move.l      d1,d0
           asl.l       #2,d1                   ; size X
           adda.l      #hc_len,a1
           move.l      a1,hc_curr(a0)
           subq.l      #1,d0
           moveq.l     #SIG_DFL,d2
icrl       move.l      d2,(a1)+                ; default all handlers
           dbra        d0,icrl

           move.l      a1,hc_samask(a0)
           adda.l      d1,a1
           move.w      msg.txi(a4),d2
           btst        #0,d2                   ; uval array ?
           beq.s       hnu
           move.l      a1,hc_uval(a0)
           adda.l      d1,a1
hnu        btst        #1,d2                   ; priority array ?
           beq.s       hnuex
           move.l      a1,hc_prio(a0)
           move.w      #1,hc_priid(a0)
           adda.l      d1,a1

hnuex:
* now link into list of receiving channels
           move.l      a0,ch_sigh(a5)
           st.b        ch_alloc(a5)
           move.l      dev_joblist(a3),ch_nxtch(a5)
           move.l      a5,dev_joblist(a3)
           clr.l       d0
           bra         sm_ret


* subr: a4 msg-4, a3 dev_df a5 ch_def d7 flag -> copy old event to (a1),d6 len
* signr==0 : cancel previous event
insert_event:
           tst.w       msg.txi+ml(a4)          ; unit?
           beq.s       cticks
           move.w      ticks_per_second,d0
           move.l      msg.t_ticks+ml(a4),d1   
           bsr         mul16x32
           move.l      d1,msg.t_ticks+ml(a4)
           move.l      msg.t_int+ml(a4),d1
           bsr         mul16x32
           move.l      d1,msg.t_int+ml(a4)
cticks:    
           move.l      dev_gticks(a3),d0
           add.l       d0,msg.t_ticks+ml(a4)   ; form now on
           move.l      msg.t_evid+ml(a4),d1
           lea         ch_events(a5),a2        ; 'previous'
iltev:     move.l      (a2),d0
           beq.s       insrt
           move.l      d0,a0                   ; search for old entry
           cmp.l       msg.t_evid+ml(a0),d1    ; event id
           beq.s       rmv_oldevent
           lea         ml.nxt(a0),a2
           bra.s       iltev

rmv_oldevent:
           move.l      ml.nxt(a0),(a2)         ; unlink
           tst.b       d7
           beq.s       nmcp                    ; no response wanted
           lea         ml(a0),a2
           move.l      dev_gticks(a3),d1
           sub.l       d1,msg.t_ticks(a2)      ; remaining time
           tst.w       msg.txi(a2)
           beq.s       nocnv
           move.w      ticks_per_second,d2
           move.l      msg.t_ticks(a2),d1      ; convert rest time
           bsr.s       udiv32x16
           move.l      d1,msg.t_ticks(a2)
*          tst.l       d1
*          bne.s       cnvnz
*          moveq       #1,d1                   ; must be at least '1'
*cnvnz:    move.l      d1,msg.t_ticks(a2)
*          move.l      msg.t_int(a2),d1
*          bsr.s       udiv32x16
*          move.l      d1,msg.t_int(a2)

nocnv:     tst.l       msg.t_ticks(a2)         ; must return at least 1 unit
           bne.s       nznc
           addq.l      #1,msg.t_ticks(a2)
nznc:
           move.l      msg.len(a2),d6
           subq.l      #1,d6
           move.l      #'%MSG',msg.magic(a2)   ; return with correct magic
cpml:      move.b      (a2)+,(a1)+ 
           dbra        d6,cpml                 ; and copy return msg into buffer
           move.l      msg.len+ml(a0),d6
nmcp:      move.w      mm.rechp,a2
           move.l      a3,-(a7)
           jsr         (a2)                    ; free mem
           move.l      (a7)+,a3

insrt:     tst.l       msg.signr+ml(a4)        ; BUGFIX, insrt label moved
           beq.s       ievret                  ; return, was a cancel/info request

           lea         ch_events(a5),a2        ; now insert into time sorted list
           move.l      msg.t_ticks+ml(a4),d1
istev:     move.l      (a2),d0
           beq.s       link_in
           move.l      d0,a0
           cmp.l       msg.t_ticks+ml(a0),d1
           bcs.s       link_in
           lea         ml.nxt(a0),a2
           bra.s       istev

link_in:   move.l      a4,(a2)
           move.l      d0,ml.nxt(a4)           ; next event pointer
           clr.l       ml+msg.magic(a4)        ; clear magic to keep
ievret     rts                                 ; clean heap entry 


mul16x32:
           move.l      d1,d2
           clr.w       d2
           swap        d2
           mulu        d0,d2
           swap        d2                      ; ignore overflow
           clr.w       d2
           mulu        d0,d1
           add.l       d2,d1
           rts

*div16x32: 
*          divu        d0,d1
*          andi.l      #$ffff,d1
*          rts

* unsigned div d1/d2; 32/16
* result ->d1, rest ->d0
udiv32x16:
           MOVEQ       #$00,D0
           SWAP        D1
           MOVE.W      D1,D0
           DIVU        d2,D0                   * MSW/d2
           SWAP        D0                      * rest
           MOVE.W      D0,D1                   *      + LSW -> d1
           SWAP        D1
           DIVU        d2,D1
           MOVE.W      D1,D0                   * division result -> d0
           SWAP        D1
           EXG         D0,D1       <-->        * rest -> d1
           RTS




* a3 device def. block 
find_next_event:
           suba.l      a1,a1
           moveq.l     #-1,d0                  ; unisgned, max tick count
           moveq.l     #0,d4
           move.l      dev_joblist(a3),a2
           move.l      a2,d3
           beq.s       fnele
fnel1:
           move.l      ch_events(a2),a0
           cmpa.l      d4,a0
           beq.s       fnel1nx                 ; no events for this channel
           move.l      msg.t_ticks+ml(a0),d1
           cmp.l       d0,d1
           bcs.s       fnnm
fnel1nx:   move.l      ch_nxtch(a2),d3
           beq.s       fnele
           move.l      d3,a2
           bra.s       fnel1

fnnm:      move.l      a0,a1                   ; new ticks-min
           move.l      d1,d0
           bra.s       fnel1nx

fnele:     move.l      a1,dev_nxevent(a3)
           sub.l       dev_gticks(a3),d0
           bcc.s       fnnz                    ; >=0
           clr.l       d0
fnnz:      move.l      d0,dev_nticks(a3)
           rts
***
getjobid:
           move.l      a4,-(a7)
           move.l      sv_jbpnt(a6),a4
           move.l      a4,d1
           sub.l       sv_jbbas(a6),d1
           asr.l       #2,d1
           swap        d1
           move.l      (a4),a4
           move.w      job.tag(a4),d1
           swap        d1
           move.l      (a7)+,a4
           rts
*****
* d1=jobid, d2.l=interruptnr, d3=tmout, a1= priority, a2=usr_value
*
* verify jobid, find handler, send signal(s)
send_sig:  
           move.l      a2,d7                   ; save uval
           lea         -tmp_size(a7),a7
           move.l      a7,a2

           tst.l       d1
           bpl.s       checkjob                ;
           bsr.s       getjobid                ; remove when removing $2f

checkjob:
           movem.l     d1/a0-a1,-(a7)          ; d1=jobid
           move.l      d1,d4
           beq         ix_bj                   ; don't send signal to job 0
           bsr         getjbasea0
           tst.w       d0
           bne         ix_bj
                                                           
check_d3:
           tst.l       d3
           beq.s       oj                      ; is it current job ?

* search for handler
roj:       move.l      a0,a5
           bsr.s       fjhand
           exg.l       a5,a0                   ; a5 chdef, a0 jobbase
           tst.l       d0
           beq         ix_nf
           bra.s       found
oj:
           move.l      sv_jbpnt(a6),a4         ; BUGFIX ??
           cmpa.l      (a4),a0
           bne.s       roj
           tst.l       d2
           beq.s       roj                     ; only testing job & handler
           bra         ix_nc
***
* a5 chdef, sets tmp.imode, tmp.force
chk_ssc:
           clr.w       tmp.imode(a2)
           cmpi.l      #'%CSG',h_type(a4)
           bne.s       rnss
           move.l      hc_scinf(a4),d0
           beq.s       rnss
           move.l      d0,a1
           move.w      sce.stat(a1),d0         ; ATOM=IM_BLOCK & NOATM=IM_EXP
           beq.s       rssc
* these signals get special treatment
           cmp.b       #SIGSEGV,d5 ; signr<=32!, lazy eval
           beq.s       cssf                    ; hard exception
           cmp.b       #SIGILL,d5
           beq.s       cssf
           cmp.b       #SIGBUS,d5
           bne.s       nelj
cssf       st.b        tmp.force(a2)
           moveq       #IM_LJMP,d0
           lea         sce.cntxt(a1),a3
           move.l      a3,tmp.haddr(a2)        ; longjmp ctxt
rssc       move.w      d0,tmp.imode(a2)
rnss       rts

nelj:
           cmpi.b      #IM_BLOCK,d0
           beq.s       rssc
           addq.w      #1,ch_ispf(a5)          ; set expire flag
           bra.s       rssc


***
* find jobs receiving channel,a3  devdriver, d1 jobid
* returns a0 channel def block, d0<> if successfull
fjhand:
           move.l      dev_joblist(a3),d0
fjlll      beq.s       fjret                   ; not found
           move.l      d0,a0
           cmp.l       ch_ojob(a0),d1
           beq.s       fjret                   ; found
           move.l      ch_nxtch(a0),d0
           bne.s       fjlll
fjret      rts



***
* rearange registers
* from now: d6 priority, d4 jobid, d5 signr, a5 chdefbl
* later a0 jobbase
found:
           move.l      a1,d6                   ; save priority
           move.l      d1,d4                   ; jobid
           move.l      d2,d5                   ; signalnr

           clr.w       tmp.force(a2)
           move.l      ch_sigh(a5),a4

* see whether emulated syscall in progress and set flags
           
           bsr         chk_ssc

* now handle the different jobhandlers
fnss:
           cmp.w       #$4afc,(a4)
           bne         ix_bl
           cmp.l       #'%SIG',h_type(a4)
           bne         try_csig
***
*
* '%SIG' specific
*
           move.l      h_type(a4),tmp.htype(a2)
           tst.l       d5                      ; signr==0 
           beq.s       test_ok
           bmi         ix_ni                   ; signr<=0 ??

           move.l      d6,d0
           andi.l      #P_RES_S,d0 ; some unimplemented bits ?
           bne         ix_ni
           move.l      hs_prio(a4),tmp.prio(a2)

           tst.w       tmp.imode(a2)
           bne.s       ssic

           move.l      hs_handler(a4),tmp.haddr(a2)
           beq         ix_nc
           move.l      hs_stackmin(a4),tmp.stack(a2)

ssic       bsr         sub_penq

* do_int | set job_stat&do_int | err_nc | err_smp
* if d0=sgn_def : clr d0, just proceed
* if d0<0 : bra end, error
* if d0>=0 : do_int

           cmp.w       #sgn_def,d0
           beq.s       cd0
           tst.w       d0
           bmi         end
nxi
           movem.l     d0-d2,tmp.sregs(a2)
           bra         do_int

cd0:       clr.l       d0
           bra.s       nxi

test_ok:   clr.l       d0
           bra         end

*********************
* '%CSG' specific
* subroutine to check blocking conditions and evaluate
* handler information 
cseval:
           cmp.l       #'%CSG',h_type(a4)
           bne         cse_err.ni
           move.l      h_type(a4),tmp.htype(a2)
           move.l      hc_stack(a4),tmp.stack(a2)
           cmpi.l      #32,d5                  ; signr>32 ?
           bhi         cse_err.ni
           cmp.l       hc_nsigs(a4),d5
           bhi         cse_err.or
           tst.l       d5                      ; signr==0 (test job&handler)
           beq         csdone
           bmi         cse_err.ni              ; signr<=0 ??

           move.l      d5,d2
           asl.l       #2,d2

           cmp.w       #IM_LJMP,tmp.imode(a2)
           beq.s       ljsm
csfh:
* default/current/ignored ?
           cmp.l       #SIGKILL,d5
           beq         default                 ; ... always do default action
           move.l      hc_curr(a4),a1
           move.l      0(a1,d2.l),d0           ; current handler
           beq         ignore                  ; ignore signal
           cmp.l       #SIG_DFL,d0
           beq         default

bch:       move.l      d0,tmp.haddr(a2)

ljsm:      moveq.l     #1,d0
           move.l      d5,d2
           subq.l      #1,d2
           asl.l       d2,d0                   ; sigmask
           move.l      d0,d3                   ;   save it for a moment

           tst.w       tmp.force(a2)
           bne.s       cseforce                ; ignore any blocking

           and.l       hc_block(a4),d0         ; blocked?
           and.l       #NOBLOCKMASK,d0
           bne         is_blocked
cseforce:
* check for syscall
           move.w      tmp.imode(a2),d0
           beq.s       nssc
           move.l      hc_scinf(a4),a3
           move.l      sce.pend(a3),a3         * *pend
           or.l        d3,(a3)
           move.l      hc_uval(a4),d1
           beq.s       csneret
           move.l      d1,a3
           move.l      d5,d1
           asl.l       #2,d1
           move.l      d7,0(a3,d1.l)
           subq.w      #IM_BLOCK,d0
           beq         csdone
           bra.s       csneret

nssc       not.l       d3
           and.l       d3,hc_pend(a4)          ; delete any pending instances

csneret    clr.l       d0          ; proceed
           rts


cse_err.ni moveq       #-19,d0
           rts
cse_err.or moveq       #-4,d0
           rts

rpri:      move.l      hc_prio(a4),a1
           move.l      d5,d2
           asl.l       #2,d2
           move.l      0(a1,d2.l),tmp.prio(a2)
           bra.s       csp


* mangle tmp.prio : always apply #NORM while in emulated syscall
pmng:      moveq.l     #0,d0
           move.l      d1,d0
           andi.w      #%111,d0
           andi.l      #^(WFJOB+WFIO+SUSP),d1
           asl.w       #3,d0
           add.w       d0,d1
           asl.w       #3,d0
           add.w       d0,d1
           rts


try_csig:
           move.l      d6,d0
           andi.l      #P_RES_S,d0 ; any unimplemented bits ?
           bne         ix_ni

           cmp.l       #'%CSG',h_type(a4)
           bne         ix_ni

           tst.w       hc_priid(a4)
           bne.s       rpri
           move.l      hc_prio(a4),tmp.prio(a2)
           
csp:
           tst.w       tmp.imode(a2)
           beq.s       cscp
           move.l      tmp.prio(a2),d1
           bsr.s       pmng
           move.l      d1,tmp.prio(a2)
           move.l      d6,d1
           bsr.s       pmng
           move.l      d1,d6

cscp:      bsr         sub_penq
           tst.w       d0
           bmi         end                     ; priority too small ?
           movem.l     d0-d2,tmp.sregs(a2)     ; save results..

           bsr         cseval
           cmp.w       #-1,d0
           beq         test_ok     ; report sucess
           blt         end         ; report error

           tst.b       hc_ixx(a4)
           bne         cs_blocked              ; checksig in progress
sxcc       movem.l     tmp.sregs(a2),d0-d2
           cmp.w       #sgn_def,d0
           beq.s       cs_def
           tst.w       d0
           bgt.s       cs_sstat
cs_doint:  
           bra         do_int

cs_sstat:
           tst.b       hc_defer(a4)
           beq.s       cs_doint
           bsr         sub_chaddr              ; test handler address
           cmp.w       #-1,d0
           beq         test_ok
           blt         end                     ; error
* now reuse frame of defered signal:
*   doesn't need special handling for INSYSCALL since nothing gets deferred
*   while INSYSCALL
           clr.b       hc_defer(a4)
           move.l      job.a7(a0),a1
           move.l      tmp.haddr(a2),(a1)      ; substitute new handler
           lea         4(a1),a1                ; addr of frame to be reused
           movem.l     tmp.sregs(a2),d0-d2
           move.w      d0,svb.sgn(a1)
           move.l      d1,svb.fs1(a1)
           move.l      d2,svb.fs2(a1)

           move.l      esvb.signr(a1),d2       ; get old values
           move.l      esvb.uval(a1),d0

           move.l      tmp.prio(a2),job.d5(a0)
           move.l      d5,job.d4(a0)
           move.l      d7,job.a4(a0)
           move.l      d5,esvb.signr(a1)
           move.l      d7,esvb.uval(a1)
           move.l      d2,d5
           move.l      d0,d7
           bsr.s       is_blocked              ; mark old as pending
           movem.l     tmp.sregs(a2),d0-d2
           bsr         sub_sysci
           bra         test_ok


cs_def:
           clr.l       d0
           tst.w       tmp.imode(a2)           
           bne         cs_doint                ; don't set defer flag!

           tas.b       hc_defer(a4)            ; defer max one
           bne.s       cs_blocked
           movem.l     d0-d2,tmp.sregs(a2)
           bra         cs_doint


cs_blocked:
           bsr.s       is_blocked
           bra         test_ok


suval:     move.l      d0,a3
           move.l      d5,d0
           asl.l       #2,d0
           move.l      d7,0(a3,d0.l)
           bra.s       ib1

is_blocked:
           move.l      hc_uval(a4),d0
           bne.s       suval
ib1:       moveq.l     #1,d0
           move.l      d5,d2
           sub.l       d0,d2
           asl.l       d2,d0
           or.l        d0,hc_pend(a4)          ; mark pending
           bra.s       csdone                  ; no further action needed

default:   move.l      hc_def(a4),d1
           beq.s       sysdefl
           move.l      d1,a1
           move.l      0(a1,d2.l),d0
           beq.s       ignore                  ; SIG_IGN
           bra         bch

sysdefl:   moveq       #SIG_DFL,d0
           bra         bch


ignore:    moveq.l     #1,d0
           sub.l       d0,d5                   ;  reset pending
           asl.l       d5,d0                   ; sigmask
           not.l       d0
           and.l       d0,hc_pend(a4)
csdone:    moveq.l     #-1,d0
           rts




************************
* now do the software interrupt
*

* subr, alloc stack frame and store context info on stack
setupframe:
           move.l      job.a7(a0),a1
           cmp.l       #'%CSG',tmp.htype(a2)
           bne.s       simpleframe
           suba.l      #esvb_len,a1
           bra.s       cofr

simpleframe:
           suba.l      #svb_len,a1 ; adjust to frame
cofr:      cmpa.l      tmp.stack(a2),a1        ; enough stack for interrupt?
           bcs.s       err.nostack ; blt
           move.w      d0,svb.sgn(a1)          ; set frame type
           move.l      d1,svb.fs1(a1)
           move.l      d2,svb.fs2(a1)
           rts

err.nostack:
           addq.l      #4,a7
           moveq       #-3,d0                  ; err.om
           bra         end

sub_sysci: cmp.w       #sgn_wfj,d0
           bne.s       cssci
           move.l      d2,a3
           clr.b       job.wflag(a3)
cssci:     move.w      #1,job.stat(a0)
           rts

sub_chaddr:
           move.l      tmp.haddr(a2),d3        ; extra check for ignore/default
           beq.s       cha_ign                 ; ->ignore
           moveq       #1,d0
           cmp.l       d0,d3                   ; =1 ?
           beq.s       disysd                  ; -> SIG_DFL
           and.l       d0,d3                   ; uneven address ?
           bne.s       cha_bl
           bra.s       cha_ret
* do system default action
disysd:    cmp.b       sysdeftab,d5
           bcc.s       cha_ign
           lea         sysdeftab,a3
           move.b      0(a3,d5.w),d0
           beq.s       cha_ign
dik:       lea         killjob,a3
           move.l      a3,tmp.haddr(a2)
cha_ret:   clr.l       d0
           rts
cha_ign:   moveq.l     #ERR.NC,d0
           rts
cha_bl:    moveq.l     #ERR.BL,d0
           rts

* main entry point
do_int:
           tst.w       tmp.imode(a2)
           bne.s       diisysc                 ; insides syscall special treatment
           bsr.s       sub_chaddr              ; check handler address
           cmp.w       #-1,d0
           beq.s       eret                    ; ignore
           blt.s       end                     ; other error

           movem.l     tmp.sregs(a2),d0-d2
           bsr         setupframe
           tst.w       d0
           ble.s       csdi
           bsr.s       sub_sysci               ; interrupt  system call
csdi:
           move.l      job.pc(a0),svb.ret(a1)  ; save PC
           move.w      job.sr(a0),svb.sr(a1)               ; save SR
           move.l      job.a7(a0),svb.a7(a1)               ; for REQSIGSTACK
           move.l      job.d4(a0),svb.d4(a1)               ; save D4-D6
           move.l      job.d5(a0),svb.d5(a1)
           move.l      job.d6(a0),svb.d6(a1)
           move.l      job.a4(a0),svb.a4(a1)
           move.l      job.a5(a0),svb.a5(a1)               ;  .... A4/A5

           move.l      d5,job.d4(a0)                       ; interrupt number
           move.l      d7,job.a4(a0)                       ; usr val
           move.l      d6,job.d5(a0)                       ; priority

           cmp.l       #'%CSG',tmp.htype(a2)
           beq.s       csg_setpc

           lea         int_ret,a3
           move.l      a3,svb.finish(a1)
           move.l      a1,job.a7(a0)                       ; adjust jobs SP
           move.l      tmp.haddr(a2),job.pc(a0)            ; handler routine
eret:      
           clr.l       d0
end:       
           movem.l     (a7)+,d1/a0-a1
           lea         tmp_size(a7),a7
           rts

diisysc:
           movem.l     tmp.sregs(a2),d0-d2
           bsr         sub_sysci
           cmp.w       #IM_LJMP,tmp.imode(a2)
           bne.s       eret
* now arrange for the longjmp(), cntxt addr in tmp.haddr
           move.l      tmp.haddr(a2),job.a4(a0)            ; pass pointer to jmpbuf
           lea         do_ljmp,a3
           move.l      a3,job.pc(a0)
           move.l      d7,job.d4(a0)
           bra.s       eret


csg_setpc: 
           lea         cint_ret,a3
           move.l      a3,svb.finish(a1)       ; cleanuproutine address
           move.l      a4,esvb.hstruct(a1)     ; some more info
           move.l      d5,esvb.signr(a1)       ;  for handler restore
           move.l      d7,esvb.uval(a1)        ;  and reset

           move.l      tmp.haddr(a2),-(a1)
           move.l      a1,job.a7(a0)                       ; adjust SP
           lea         ccal_setup,a3
           move.l      a3,job.pc(a0)           ; wrapper for c68 calls

           clr.l       d0
srt:       bra.s       end


ix_nf      moveq       #ERR.NF,d0
           bra.s       srt
ix_bl:     moveq       #ERR.BL,d0
           bra.s       srt
ix_ni      moveq       #ERR.NI,d0
           bra.s       srt
ix_bj      moveq       #ERR.NJ,d0
           bra.s       srt
ix_or      moveq       #ERR.OR,d0
           bra.s       srt
ix_nc      moveq       #ERR.NC,d0
           bra.s       srt




*****************
* sub_penq : check if job can be interrupted
*   returns d0, d1=svb.sf1, d2=svb.sf2
*           d0 = sgn_def : defer after syscall
*           d0 = -1       : err_nc
*           d0= -20       : error, priority to small
*           d0 >= 0       : interrupt syscall

sysc_def   equ         $1000

sub_penq:
           move.w      job.sr(a0),d0
           andi.w      #SUPER,d0               ; job in sv mode?
           bne.s       ret_nc                  ; must wait

           move.l      tmp.prio(a2),d3         ; required priority
           andi.l      #P_RES_R,d3 ; some unimplemented features ?
           bne.s       ret_ni
           move.l      tmp.prio(a2),d3
           move.l      d6,d0                   ; priority of interrupt
           
           tst.w       job.stat(a0)
           bne         stnook                  ; test some conditions

           andi.w      #NORM,d3
           andi.w      #NORM,d0
           cmp.w       d3,d0
           ble         ret_smp                 ; priority too small
           clr.l       d0
           clr.l       d1
           clr.l       d2
           rts

ret_nc:    moveq.l     #-1,d0
           rts
ret_ni:    moveq.l     #-19,d0
           rts

wj:        andi.w      #WFJOB,d0
           andi.w      #WFJOB,d3
           cmp.w       d3,d0
           bgt.s       wji
           move.l      tmp.prio(a2),d3         ; required priority to signal
           move.w      d6,d0
           move.w      d6,d1
           andi.w      #DF_WJOB,d1
           bne         ret_def                 ; defer signal after job released
           bra         ret_smp

* deal with jobs waiting for another job
* first find that other job:
wji:       move.l      sv_jbbas(a6),a1                     ; base of jobtab
           move.l      sv_jbtop(a6),d3                     ; top
jslo       cmp.l       a1,d3
           bcs.s       ret_errbj
           move.l      (a1)+,a3
           move.l      a3,d0
           bmi.s       jslo
           tst.b       job.wflag(a3)
           beq.s       jslo
           cmp.l       job.wjob(a3),d4
           beq.s       jobfound
           bra.s       jslo

ret_errbj: moveq.l     #ERR.NJ,d0
           rts                                 ; shouldn't ever happen

jobfound:
           move.l      a1,d0                   ; compute jobtable index
           sub.l       sv_jbbas(a6),d0
           asr.l       #2,d0
           subq.w      #1,d0                   ;
           swap        d0
           move.w      job.tag(a3),d0
           swap        d0                      ;    to get jobid

* SAVE STATE INFO
           move.l      d0,d1                   ; save jobid
           move.l      a3,d2
           moveq.l     #sgn_wfj,d0
           rts

t3:        andi.w      #WFIO,d0
           andi.w      #WFIO,d3
           cmp.w       d3,d0       
           bgt.s       schan
           move.l      tmp.prio(a2),d3
           move.w      d6,d1
           move.w      d6,d0
           andi.w      #DF_WIO,d1              ; defer signal after completion 
           bne         ret_def
           bra         ret_smp


* try to figure out why job.stat is set and handle it
stnook:    
           cmp.w       #-2,job.stat(a0)        ; waiting for another job ?
           beq         wj                      ; ... check if we should signal

* ... assume WFIO and try to verify
* by examining a0

schan:     move.l      sv_chbas(a6),a1
           move.w      2+job.a0(a0),d0         ; chnr
           cmp.w       sv_chmax(a6),d0
           bhi.s       suspcheck
           asl.w       #2,d0                   ; *4
           adda.w      d0,a1
           tst.b       (a1)
           bmi.s       suspcheck
           move.l      (a1),a3                 ; ch.defblock  -> a3
           move.w      job.a0(a0),d0
           cmp.w       ch_tag(a3),d0           ; another channel with same
           bne.s       suspcheck               ; index in chan table?

           tst.b       ch_stat(a3)
           bge.s       suspcheck
           cmp.l       ch_jobwt(a3),d4         ; jobid matches?
           bne.s       suspcheck
cnfound:   
           move.l      tmp.prio(a2),d3
           move.l      d6,d0
           move.l      d6,d1
           andi.w      #WFIO,d0
           andi.w      #WFIO,d3
           cmp.w       d3,d0
           bgt.s       t3_int
           move.w      d6,d1
           andi.w      #DF_WIO,d1              ; can it be defered ?
           bne.s       ret_def
           bra.s       ret_smp

t3_int:
           moveq.l     #sgn_ios,d0
           move.l      job.a0(a0),d1           ; save chid
           clr.l       d2
           move.b      ch_actn(a3),d2          ; save action
           rts


suspcheck:
           move.l      d6,d0                   ; send priority
           move.l      d6,d1

           move.l      tmp.prio(a2),d3         ; required priority
           andi.w      #SUSP,d0                
           andi.w      #SUSP,d3                
           cmp.w       d3,d0
           bgt.s       int_susp    
           move.w      d6,d0
           andi.w      #DF_SUSP,d1
           bne.s       ret_def                 ; defer signal after syscall
           bra.s       ret_smp

           
int_susp:
           moveq.l     #sgn_sus,d0
           clr.l       d1
           clr.l       d2
           rts

ret_smp:
           moveq.l     #-20,d0
           rts

ret_def:   move.l      tmp.prio(a2),d3
           move.l      d6,d0                   ; priority of interrupt
           andi.w      #NORM,d3
           andi.w      #NORM,d0
           cmp.w       d3,d0
           ble         ret_smp                 ; too small even to defer
           moveq.l     #sgn_def,d0
           clr.l       d1
           clr.l       d2
           rts




***********************************************************************
* interrupt setup, cleanup and longjmp routines
*
* standard cleanup, restores jobs stack, registers and sr flags
* tested only on 68008

*int_ret:
*          lea         -4(a7),a7
*_iret_:   trap        #0
*          move.l      USP,a5
*          move.l      svb.a7(a5),a4
*          move.l      a4,USP
** if not g.o. 68000: move.w       #0,(-a7)  * exception format word
*          move.l      svb.ret(a5),-(a7)
*          move.w      svb.sr(a5),-(a7)
*          move.l      svb.d4(a5),d4
*          move.l      svb.d5(a5),d5
*          move.l      svb.d6(a5),d6
*          move.l      svb.a4(a5),a4
*          move.l      svb.a5(a5),a5
*          rte

* old version, can't relocate stack, but then
* it should work on all 680x0 chips
int_ret:
           lea         -4(a7),a7
           move.l      svb.a5(a7),a5
           move.l      svb.a4(a7),a4
           move.l      svb.d6(a7),d6
           move.l      svb.d5(a7),d5
           move.l      svb.d4(a7),d4
           lea         svb.sr(a7),a7
           rtr                     ; ccr restored, sr must be preserved

** return from C68 signal handler
cint_ret:
           lea         -4(a7),a7               ; adjust esvb frame
* restore mask&handler
           move.l      esvb.hstruct(a7),a1
           tst.w       hc_activ(a1)            ; sigaction() called 
           beq.s       noreset                 ; .. in the meantime ?

           subq.w      #1,hc_activ(a1)
           move.l      esvb.oblock(a7),hc_block(a1)        ; restore mask
           move.l      hc_curr(a1),a0
           move.l      esvb.signr(a7),d0
           asl.l       #2,d0
           move.l      esvb.ohand(a7),0(a0,d0.l)           ; restore handler
           
noreset:
           cmp.w       #sgn_raise,svb.sgn(a7)  ; gets raised anyway
           beq.s       ncs                                 ; no need for recursion
           movem.l     d3/a2/a3/d7/a5,-(a7)
           move.l      a1,a4                               ; handler pointer
           bsr         _checksig
           movem.l     (a7)+,d3/a2/a3/d7/a5

ncs:       movem.l     esvb.d0(a7),d0-d2/a0/a1 ; restore scratch regs
**         bra.s       _iret_      ; requires stack relocation, not yet
** tested with all 680x0 chips, so use this hack instead:
           move.l      svb.a5(a7),a5
           move.l      svb.a4(a7),a4
           move.l      svb.d6(a7),d6
           move.l      svb.d5(a7),d5
           move.l      svb.d4(a7),d4
           move.w      svb.sr(a7),esvb.sr(a7)
           move.l      svb.ret(a7),esvb.ret(a7)
           lea         esvb.sr(a7),a7
           rtr                     ; ccr restored, sr must be preserved

** wrapper to allow c-handlers
* first 2 instructions critical
ccal_setup:
           move.l      esvb.hstruct+4(a7),a5
           clr.b       hc_defer(a5)
* now save compiler scratch regs
           movem.l     d0-d2/a0/a1,esvb.d0+4(a7)
           move.l      (a7)+,a0                            ; handler address
           move.l      a7,a1                               ; esvb
           move.l      a4,-(a7)                            ; usrval
           move.l      d5,-(a7)                            ; priority
           move.l      d4,-(a7)                            ; signr
           move.l      a5,a4                               ; ptr to %CSG struct

           move.l      d4,d0                               ; signr
           asl.l       #2,d0                               ;   *size

           addq.w      #1,hc_activ(a4)                     ; handling in progress
           move.l      hc_block(a4),esvb.oblock(a1)        ; save procmask
           move.l      hc_curr(a4),a5
           move.l      0(a5,d0.l),esvb.ohand(a1)           ; save handler
           move.l      hc_samask(a4),a5
           move.l      0(a5,d0.l),d1
           or.l        d1,hc_block(a4)                     ; new procmask

           moveq.l     #1,d2
           sub.l       d2,d4
           asl.l       d4,d2                               ; sigmask
           and.l       hc_reset(a4),d2                     ; reset handler?
           beq.s       cssc
           lea         hc_curr(a4),a5
           move.l      #SIG_DFL,0(a5,d0.l)                 ; reset it
cssc:
           jsr         (a0)        ; call handler
           lea         12(a7),a7
           rts

* a4 -> lonjmp context, d4 uval
do_ljmp:
           move.l      a4,a0
           move.l      d4,d0
           movem.l     (a0)+,a3-a7
           move.l      (a0),a0
           jmp         (a0)


*******************************************
* system default table used by sub_chaddr
* known actions are
* 0= ignore, 1= kill
* signals>15 are ingored by default

sysdeftab: dc.b        15          ; table length
           dc.b        1           ; SIGHUP
           dc.b        1           ; SIGINT
           dc.b        1           ; SIGQUIT 
           dc.b        1           ; SIGILL 
           dc.b        1           ; SIGPIPE
           dc.b        1           ; SIGSEGV  
           dc.b        1           ; SIGBUS   
           dc.b        0           ; SIGFPE 
           dc.b        1           ; SIGKILL
           dc.b        1           ; SIGALRM
           dc.b        1           ; SIGABRT 
           dc.b        1           ; SIGTRACE 
           dc.b        0           ; SIGWINCHD 
           dc.b        0           ; SIGWMREQ 
           dc.b        1           ; SIGTERM 
           dc.w        0


* default handler special cases

killjob:   moveq.l     #5,d0
           moveq.l     #-1,d1
           moveq.l     #EINTR,d3
           trap        #1
           rts

*******************************************

**** c-interface 
** parameter passing through stack

* sgv(chid,0)                      : sigcleanup()
* sgv(chid,1,signr,oadr,nadr)      : sigaction(signr,oadr,nadr)
* sgv(chid,2,how,&omask,&mask)     : sigprocmask(how,oset,nset), c/e block mask
* sgv(chid,3,&mask)                : sigpending() examine pending
* sgv(chid,4,signr,uval)           : raise() signr,uval
* sgv(chid,5,signr,uval)           : fraise()
* sgv(chid,6)                      : checksig(), raise pending signals
* sgv(chid,7,signr,&oprio,&prio)   : c/e priority
* sgv(chid,8,flags,&sctl,&sycfunc,argc,..)     
*    flags: none yet               : run sysfunc as emulated system call,
*    struct {pp;rval;} sctl        : return qdos error, actual return vale
*                                  : in sctl->rval
* sgv(chid,9,flags)                : syscontrol, set flags & return pflag addr
*
*

MAXSIGV    equ         9
PAR1       equ         12+52
PAR2       equ         16+52
PAR3       equ         20+52
PAR4       equ         PAR3+4

c_sigvec:
           movem.l     d1-d7/a0-a5,-(a7)       ; (7+6)*4 = 52
           move.l      52+8(a7),d4 ; opcode
           bmi         svnoimp
           cmp.w       #MAXSIGV,d4
           bgt         svnoimp

           moveq       #sms.info,d0
           trap        #1

svc:       move.l      52+4(a7),d3 ; channel   
           move.l      d3,d2

           move.w      sr,d7
           trap        #0
           cmp.w       sv_chmax(a0),d1
           bhi.s       su_err.no
           asl.l       #2,d3
           move.l      sv_chbas(a0),a0
           adda.w      d3,a0
           tst.b       (a0)
           bmi.s       su_err.no
           tst.b       (a0)
           bmi.s       su_err.no
           move.l      (a0),a0

           move.l      ch_driver(a0),a3
           lea         io,a2
           cmp.l       4(a3),a2                ; is it really a signal_ channel?
           bne.s       su_err.bp
           cmp.l       ch_ojob(a0),d1          ; is it our channel ?
           beq.s       svfch

           lea         0-$18(a3),a3            ;  ... serach it
           bsr         fjhand
           tst.l       d0
           beq.s       su_err.no

svfch:     move.l      ch_sigh(a0),a4
           cmpi.l      #'%CSG',h_type(a4)
           bne.s       su_err.bp

* set scheduler flag; as it is in the ch def block it can be done in
* supervisor mode only - doing it here saves from doing all consistency
* checks twice. The scheduler will sort needs to check the flag anyway

           addq.w      #1,ch_ispf(a0)

           move.w      d7,sr                   ; return to normal state
           move.l      a0,a5

           asl.l       #2,d4
           jmp         table(pc,d4.l)

su_err.no  move.w      d7,sr
           bra         sverr.no    
su_err.bp  move.w      d7,sr
           bra         sverr.bp


* a1 chdefblock
table:     bra         sigcleanup
           bra         sigact
           bra         sigprocmask
           bra         sigpending
           bra         raise
           bra         fraise
           bra         checksig
           bra         chprio
           bra         dosyscall
           bra         sysctl


* sgv(chid,8,flags,&sctl,&sycfunc,argc,..)     
*            PAR1   PAR2  PAR3    PAR4
*                                  : run sysfunc as emulated system call,
*                                  : return qdos error
*          flags:      0           : none so far
*          struct sctl{pend,rv}    : pflag and return value
sctl.ppend equ         0
sctl.rtval equ         4

dosyscall:
* alloc and link in sysce struct 
           move.l      PAR1(a7),d5 ; any flags ?
           bne         sverr.ni                ; sorry, no flags so far

           cmpa.l      hc_stack(a4),a7
           bcs         sverr.om

           move.l      hc_scinf(a4),d0
           lea         -sce_size(a7),a7
           move.l      a7,a1

           move.l      d0,sce.nx(a1)           ; save link

           move.l      sce_size+PAR2(a7),a2
           clr.l       (a2)
           move.l      a2,sce.pend(a1)         ; set pflag addr

* set longjmp context
           lea         dsljret,a3
           move.l      a3,sce.pc(a1)
           move.l      a1,a3
           movem.l     a3/a4/a5/a6/a7,sce.cntxt(a1)

* link in
           move.w      #SCS_ATOM,sce.stat(a3)
           move.l      a3,hc_scinf(a4)

* copy args, save registers for call
           move.l      sce_size+PAR3(a7),a0    ; sysfunc()
           lea         sce_size+PAR4(a7),a1
           move.l      (a1)+,d0
           move.l      d0,d1
           asl.l       #2,d1
           suba.l      d1,a7
           move.l      a7,a2
           subq.l      #1,d0
dscal      move.l      (a1)+,(a2)+
           dbra        d0,dscal

* set syscall status, pending

           move.l      sce.pend(a3),a2
           move.l      sce.nx(a3),d0
           beq.s       pns                     ; nothing relevant pending
           move.l      d0,a1
           move.l      sce.pend(a1),a1
           move.l      (a1),d0
pns:       or.l        d0,(a2)

* call syscfunc, save retval
           jsr         (a0)
           move.l      sce.a7(a3),a7
dsljret:   move.l      sce_size+PAR2(a7),a1    
           move.l      d0,sctl.rtval(a1)       ; save return value(or uval)


* restore some regs
           move.l      a5,a0
           move.l      a3,a1
           move.l      ch_sigh(a5),a4

* get pending
           move.l      sce.pend(a1),a2
           move.l      (a2),d4
           andi.l      #^_SIGFTX,d4

           move.l      sce.nx(a1),hc_scinf(a4) ; unlink, set flags!
           lea         sce_size(a7),a7                     ; dealloc sce struct

* pass pending sigs back to user program or higher level call
* except fatal signals

           beq.s       dsnn
           move.l      hc_scinf(a4),a1

           move.l      sce.pend(a1),a2
           or.l        d4,(a2)
           bra.s       dret
dsnn:
           or.l        d4,hc_pend(a4)
* return QDOS error
dret       bra         checksig                ; raise pending signals, return 0


* sgv(chid,9,flags)                : syscontrol, set flags, return pflag addr
*             PAR1                 :    or error code
*          flags: expire  0x1
sysctl:
           move.l      hc_scinf(a4),d0
           beq         sverr.bp                ; not inside syscall
           move.l      d0,a1
           tst.w       sce.stat(a1)
           beq         sverr.bp
           move.l      sce.pend(a1),d0         ; return value
           move.l      PAR1(a7),d1
           subq.l      #1,d1
           bhi         sverr.ni
           bne.s       dcstat
           move.w      #SCS_EXP,sce.stat(a1)
*          move.l      d0,a2
*          tst.l       (a2)                    ; anything pending ?
*          bne.s       dcssf
           bra         svret

*dcssf     addq.w      #1,ch_ispf(a0)          ; set scheduler flag
*          bra         svret

dcstat     move.w      #SCS_ATOM,sce.stat(a1)
           bra         svret
           

* sgv(chid,7,signr,&oprio,&prio)   : c/e priority
*            PAR1   PAR2   PAR3
chprio:
           move.l      PAR1(a7),d5
           beq.s       scp
           tst.w       hc_priid(a4)
           beq         sverr.bp                ; non matching cardinality
           cmp.l       hc_nsigs(a4),d5
           bcc         sverr.or
           asl.l       #2,d5
           move.l      PAR2(a7),d0
           bne.s       scsp
           move.l      d0,a1
           move.l      hc_prio(a4),a2
           move.l      0(a2,d5.l),(a1)         ; set oprio
scsp       move.l      PAR3(a7),d0
           bne         svzret                  ; done
           move.l      d0,a1
           move.l      (a1),d0
           andi.l      #P_RES_R,d0
           bne         sverr.xp
           move.l      (a1),0(a2,d5.l)
           bra         svzret

scp        tst.w       hc_priid(a4)
           bne         sverr.bp                ; non matching cardinality
           move.l      PAR2(a7),d0
           beq.s       scpno
           move.l      d0,a1
           move.l      hc_prio(a4),(a1)        ; set oprio
scpno      move.l      PAR3(a7),d0
           beq         svzret
           move.l      d0,a1
           move.l      (a1),d0
           andi.l      #P_RES_R,d0
           bne         sverr.xp
           move.l      (a1),hc_prio(a4)
           bra         svzret


sigcleanup:
***        clr.b       hc_defer(a4)
***        clr.b       hc_ixx(a4)
           clr.w       hc_defer(a4)            ; instead of above
           clr.w       hc_activ(a4)
           bra         checksig

SV_RESETHAND equ       1
sigact:
           move.l      PAR1(a7),d0
           move.l      d0,d1
           cmp.l       hc_nsigs(a4),d0
           bhi         sverr.or
           move.l      PAR3(a7),d2 ; old value of sigaction
           beq.s       svnact
           move.l      d2,a2
           asl.l       #2,d0
           move.l      hc_curr(a4),a3
           move.l      0(a3,d0.l),(a2)         ; sa_handler
           move.l      hc_samask(a4),a3
           move.l      0(a3,d0.l),4(a2)        ; sa_mask
           clr.l       8(a2)                   ; sa_flags
           moveq.l     #1,d3
           sub.l       d3,d1
           asl.l       d1,d3
           and.l       hc_reset(a4),d3
           beq.s       svnact
           move.l      #SV_RESETHAND,8(a2)     ; set reset handler flag
svnact     move.l      PAR2(a7),d2
           beq.s       sigvecret
           clr.w       hc_activ(a4)
           move.l      d2,a2
           move.l      PAR1(a7),d0
           move.l      d0,d1
           asl.l       #2,d0
           moveq.l     #1,d3
           sub.l       d3,d1
           asl.l       d1,d3
           move.l      hc_samask(a4),a3
           move.l      4(a2),0(a3,d0.l)
           not.l       d3
           and.l       d3,hc_reset(a4)
           and.l       d3,hc_block(a4)
           and.l       d3,hc_pend(a4)
           cmp.l       #4*NOCATCH,d0
           beq.s       sigvecret
           move.l      hc_curr(a4),a3
           move.l      (a2),0(a3,d0.l)
           move.l      8(a2),d0
           beq.s       sigvecret               ; no flags
           not.l       d3
           and.l       d3,hc_reset(a4)

sigvecret: and.l       #NOBLOCKMASK,hc_block(a4)
           moveq.l     #0,d0
           bra         svret

sigprocmask:
           move.l      PAR3(a7),d0
           beq.s       spm_mod
           move.l      d0,a3
           move.l      hc_block(a4),(a3)
spm_mod:   move.l      PAR2(a7),d0
           beq.s       sigvecret               ; nothing to set
           move.l      d0,a2
           move.l      PAR1(a7),d0
           beq.s       spm_block
           subq.l      #1,d0
           beq.s       spm_unblock
spm_set    move.l      (a2),hc_block(a4)
           bra.s       sigpmret
spm_block  move.l      (a2),d3
           or.l        d3,hc_block(a4)
           bra.s       sigpmret
spm_unblock
           move.l      (a2),d1
           not.l       d1
           and.l       d1,hc_block(a4)         
sigpmret:  and.l       #NOBLOCKMASK,hc_block(a4)
           moveq.l     #0,d0
           bra         checksig

sigpending:
           move.l      PAR1(a7),a2
           move.l      hc_pend(a4),(a2)
           bra         sigvecret

raise:
           move.l      PAR1(a7),d2
           move.l      d2,d5
           move.l      PAR2(a7),d7

           lea         -tmp_size(a7),a7
           move.l      a7,a2
           clr.w       tmp.force(a2)
           st.b        hc_ixx(a4)
           bsr.s       sub_raise
           lea         tmp_size(a7),a7

           move.l      d0,-(a7)
           bsr         _checksig
           move.l      (a7)+,d0

           cmp.w       #-1,d0
           beq         svzret                  ; no error, just blocked
           bra         svret

* hc_ixx must be set before entering 
sub_raise:
           bsr         chk_ssc                 ; inside syscall ?
           bsr         cseval                  ; eval handler information
           cmp.w       #-1,d0
           blt.s       srr                     ; some error
           beq.s       sr_ok                   ; blocked
           move.w      tmp.imode(a2),d0
           beq.s       nscr
           subq.w      #IM_LJMP,d0
           beq.s       rljmp
           bra.s       sr_ok

nscr       clr.b       hc_ixx(a4)

           bsr         sub_chaddr              ; check handler address
           cmp.w       #-1,d0
           beq.s       sr_ok                   ; SIG_IGN
           blt.s       srha

           move.w      sr,d0
           andi.w      #SUPER,d0
           bne.s       srj                     ; must not be in sv mode
           clr.l       d0

           lea         0-esvb_len(a7),a1
           cmpa.l      tmp.stack(a2),a1
           bcs.s       srso
           move.l      a1,a7
           move.w      #sgn_raise,svb.sgn(a7)  ; frame type
           move.l      d0,svb.fs1(a7)
           move.l      d0,svb.fs2(a7)
           move.l      a1,svb.a7(a7)
           move.w      d0,svb.sr(a7)           ; just clear it
           move.l      a4,svb.a4(a7)           ; no other registers needed
           lea         sr_ok,a3                ; this is our exit point
           move.l      a3,svb.ret(a7)
           lea         cint_ret,a3
           move.l      a3,svb.finish(a7)       ; cleanup routine

           move.l      a4,esvb.hstruct(a7)     ; some xtra info
           move.l      d5,esvb.signr(a7)       ;  for handler restore
           move.l      d7,esvb.uval(a7)        ;  and reset
           move.l      d7,a4
           move.l      d5,d4

           move.l      tmp.haddr(a2),-(a7)

           jmp         ccal_setup              ; longjmp() like return
           
sr_ok:     clr.l       d0
srr:       rts

srha:      moveq       #ERR.BL,d0
           rts
srso:      moveq       #ERR.OM,d0              ; not enough stack
           rts
srni:      moveq       #ERR.NI,d0
           rts
srj:       moveq       #ERR.NJ,d0
           rts

* longjmp needed inside simulated syscall handling
rljmp:
           move.l      tmp.haddr(a2),a4        ; context addr
           move.l      d7,d4                   ; uval
           bra         do_ljmp

fraise:
           move.l      PAR1(a7),d2
           move.l      d2,d5
           move.l      PAR2(a7),d7

           lea         -tmp_size(a7),a7
           move.l      a7,a2
           st          tmp.force(a2)                       ; force flag
           st.b        hc_ixx(a4)
           bsr         sub_raise
           lea         tmp_size(a7),a7
           bra         svret


* see whether some pending signals could be handled
checksig:
           pea         svzret      ; fall through
* subroutine
_checksig:
*should use fast check for pending signals now
           move.l      hc_block(a4),d0
           not.l       d0
           move.l      hc_pend(a4),d6          ; remember old value
           and.l       d6,d0
           beq.s       _csret                  ; nothing to do

           st.b        hc_ixx(a4)              ; inside checksig flag
**         bne.s       _icsr
           move.l      hc_nsigs(a4),d1
           moveq.l     #1,d2                   ; d2 nr to check
           move.l      d2,d3                   ; d3 mask
           move.l      d0,d4                   ; (potentialy) raiseable mask

           lea         -tmp_size(a7),a7
           move.l      a7,a2                   ; tmp.* struct
           
checkloop: st.b        hc_ixx(a4)              ; set again (in loop!)
           move.l      hc_block(a4),d0
           not.l       d0
           and.l       hc_pend(a4),d0
           and.l       d3,d0                   ; sigmask(i)
           beq.s       nxchl                   ; not this one

           move.l      d2,d5       ; prepare raise() call: signr
           clr.l       d7          ;  uval

           movem.l     d1-d4/d6/a1/a4/a5,-(a7)
           move.l      hc_uval(a4),d0
           bne.s       guval
chraise:   bsr         sub_raise
           movem.l     (a7)+,d1-d4/d6/a1/a4/a5

nxchl:     asl.l       #1,d3
           addq.l      #1,d2
           cmp.w       d2,d1
           bge.s       checkloop

           lea         tmp_size(a7),a7
_csret:    clr.b       hc_ixx(a4)
           move.l      hc_pend(a4),d0
           not.l       d6
           and.l       d6,d0
           bne.s       _checksig   ; some new signals to check
           rts

guval:     move.l      d0,a1
           move.l      d5,d0
           asl.l       #2,d0
           move.l      0(a1,d0.l),d7
           bra.s       chraise

sverr.xp   moveq       #ERR.XP,d0
           bra.s       svret
sverr.or:  moveq       #ERR.OR,d0
           bra.s       svret
sverr.bp:  moveq       #ERR.BP,d0
           bra.s       svret
sverr.ni:  
svnoimp:   moveq       #ERR.NI,d0
           bra.s       svret
sverr.no:  moveq       #ERR.NO,d0
           bra.s       svret
sverr.om   moveq       #ERR.OM,d0
           bra.s       svret

svzret:    clr.l       d0
svret:     movem.l     (a7)+,d1-d7/a0-a5
           rts


****************



*sysvar_base:
*          dc.l        0

t_p_s:
ticks_per_second:
           dc.w        50

config:
           dc.b        '<<QCFX>>01'            ; Flag,level
           dc.w        cne-cns
cns:       dc.b        'Signal Extension'      ; Name
cne:       ds.w      0
           dc.w        4
           dc.l        version                 ; Version
* ticks per seconds
           dc.b        10                      ; Item type: word
           dc.b        0                       ; Item selection Key
           dc.w        t_p_s-*                 ; Pointer to item
           dc.w        0                       ;            pre-proc
           dc.w        0                       ;            post-proc
           dc.w        tps_text-*              ;            description
           dc.w        tps_attr-*              ;            attributes
           dc.w        -1                      ; End flag

tps_text   dc.w        tpte-tpts
tpts:      dc.b        'Ticks per second'
tpte:
           ds.w      0
tps_attr:  dc.w        49,61
           dc.w        -1                      ; End Flag

           end

